# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
import inspect
import operator
from itertools import accumulate, zip_longest
from typing import Sequence, Union, Optional

import numpy as np

from ._shaped_value import ShapedValue, _indices_to_indexer, _maybe_compute_size
from .arith import ScalarValue, constant, index_cast, ConstantOp
from .tensor import compute_result_shape_reassoc_list
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir.extras.meta import region_op
from ..util import (
    _get_sym_name,
    get_user_code_loc,
    infer_mlir_type,
)
from numba_cuda_mlir._mlir._mlir_libs._mlir import register_value_caster
from numba_cuda_mlir._mlir.dialects import memref, arith, vector, builtin
from numba_cuda_mlir._mlir.dialects._ods_common import (
    get_op_result_or_op_results,
    MixedValues,
    _dispatch_mixed_values,
)
from numba_cuda_mlir._mlir.dialects.memref import (
    _is_static_int_like,
    _generated_subview,
    _is_constant_int_like,
)
from numba_cuda_mlir._mlir.dialects.memref import *
from numba_cuda_mlir._mlir.ir import (
    DenseElementsAttr,
    MemRefType,
    UnrankedMemRefType,
    ShapedType,
    IndexType,
    Type,
    Value,
    VectorType,
    SymbolTable,
    InsertionPoint,
    StridedLayoutAttr,
)

S = ShapedType.get_dynamic_size()


def __alloc(
    op_ctor,
    sizes: Sequence[Union[int, Value]],
    element_type: Type,
    memory_space=None,
    alignment=None,
    loc=None,
    ip=None,
):
    dynamic_sizes = []
    memref_shape = []
    for s in sizes:
        if isinstance(s, int):
            memref_shape.append(s)
        else:
            memref_shape.append(ShapedType.get_dynamic_size())
            dynamic_sizes.append(s)
    result_type = T.memref(*memref_shape, element_type=element_type, memory_space=memory_space)

    symbol_operands = []
    return get_op_result_or_op_results(
        op_ctor(
            result_type,
            dynamic_sizes,
            symbol_operands,
            alignment=alignment,
            loc=loc,
            ip=ip,
        )
    )


_alloc = alloc


def alloc(
    sizes: Union[int, Value],
    element_type: Type = None,
    *,
    memory_space=None,
    alignment=None,
    loc=None,
    ip=None,
):
    return __alloc(
        AllocOp,
        sizes,
        element_type,
        memory_space=memory_space,
        alignment=alignment,
        loc=loc,
        ip=ip,
    )


_alloca = alloca


def alloca(
    sizes: Union[int, Value],
    element_type: Type = None,
    memory_space=None,
    alignment=None,
    loc=None,
    ip=None,
):
    return __alloc(
        AllocaOp,
        sizes,
        element_type,
        memory_space=memory_space,
        alignment=alignment,
        loc=loc,
        ip=ip,
    )


def load(memref: Value, indices: Sequence[Union[Value, int]], *, loc=None, ip=None):
    indices = list(indices)
    for idx, i in enumerate(indices):
        if isinstance(i, int):
            indices[idx] = constant(i, index=True)
        elif isinstance(i, Value):
            if not isinstance(i.type, IndexType):
                i = index_cast(i, to=IndexType.get(), loc=loc, ip=ip)
                indices[idx] = i
        else:
            raise TypeError(f"expected {i=} to be either int or Value")
    return get_op_result_or_op_results(LoadOp(memref, indices, loc=loc, ip=ip))


def store(
    value: Value,
    memref: Value,
    indices: Sequence[Union[Value, int]],
    *,
    loc=None,
    ip=None,
):
    indices = list(indices)
    for idx, i in enumerate(indices):
        if isinstance(i, int):
            indices[idx] = constant(i, index=True)
        elif isinstance(i, Value):
            if not isinstance(i.type, IndexType):
                i = index_cast(i, to=IndexType.get(), loc=loc, ip=ip)
                indices[idx] = i
        else:
            raise TypeError(f"expected {i=} to be either int or Value")
    return get_op_result_or_op_results(StoreOp(value, memref, indices, loc=loc, ip=ip))


rank_reduce = object()


@register_value_caster(MemRefType.static_typeid)
@ShapedValue
class MemRefValue(Value):
    def __str__(self):
        return f"{self.__class__.__name__}({self.get_name()}, {self.type})"

    def __repr__(self):
        return str(self)

    def __getitem__(self, idx: tuple) -> "MemRefValue":
        loc = get_user_code_loc()

        if not self.has_rank():
            raise ValueError("only ranked memref slicing/indexing supported")

        if idx == Ellipsis or idx == slice(None):
            return self
        if isinstance(idx, tuple) and all(i == slice(None) for i in idx):
            return self
        if idx is None:
            return expand_shape(self, (0,), loc=loc)

        idx = list((idx,) if isinstance(idx, (int, ScalarValue, slice)) else idx)
        should_rank_reduce = rank_reduce in idx
        if should_rank_reduce:
            idx.remove(rank_reduce)

        for i, d in enumerate(idx):
            if isinstance(d, int):
                idx[i] = constant(d, index=True, loc=loc)

        if all(isinstance(d, ScalarValue) for d in idx) and len(idx) == len(self.shape):
            return load(self, idx, loc=loc)
        else:
            return _subview(self, tuple(idx), rank_reduce=should_rank_reduce, loc=loc)

    def __setitem__(self, idx, val):
        loc = get_user_code_loc()

        if not self.has_rank():
            raise ValueError("only ranked memref slicing/indexing supported")

        idx = list((idx,) if isinstance(idx, (ScalarValue, int, Value)) else idx)
        for i, d in enumerate(idx):
            if isinstance(d, int):
                idx[i] = constant(d, index=True, loc=loc)

        if all(isinstance(d, ScalarValue) for d in idx) and len(idx) == len(self.shape):
            if isinstance(val, (int, float)):
                # TODO: this is an unchecked conversion
                val = ScalarValue(val, dtype=self.dtype)
            if isinstance(val, ScalarValue):
                store(val, self, idx, loc=loc)
            elif isinstance(val.type, VectorType):
                return vector.StoreOp(
                    valueToStore=val,
                    base=self,
                    indices=idx,
                    loc=loc,
                )
            else:
                raise ValueError(f"unsupported value type for memref store: {val=}")
        else:
            _copy_to_subview(self, val, tuple(idx), loc=loc)

    def dim(self, idx, *, loc=None, ip=None):
        if self.has_rank():
            if isinstance(idx, Value) and isinstance(idx.owner.opview, ConstantOp):
                idx = idx.owner.opview.value.value
            if not isinstance(idx, int):
                raise TypeError(f"expected {idx=} to be an int")
            d = self.shape[idx]
            if d != S:
                return d
        if isinstance(idx, int):
            idx = constant(idx, IndexType.get())
        if not isinstance(idx.type, IndexType):
            idx = index_cast(idx, to=IndexType.get(), loc=loc, ip=ip)
        return DimOp(self, idx, loc=loc, ip=ip).result

    def dims(self, *, rank=None, loc=None, ip=None):
        if self.has_rank():
            return tuple(self.dim(i, loc=loc, ip=ip) for i in range(self.rank))
        if rank is None:
            raise ValueError("unranked dims requires rank for unranked memref")
        return tuple(self.dim(i, loc=loc, ip=ip) for i in range(rank))


def expand_shape(
    inp,
    newaxis_dims,
    *,
    loc=None,
    ip=None,
) -> MemRefValue:
    """Expand the shape of a memref.

    Insert a new axis that will appear at the `axis` position in the expanded
    memref shape.

    Args:
      inp: Input memref-like.
      axis: Position in the expanded axes where the new axis (or axes) is placed.

    Returns:
       View of `a` with the number of dimensions increased.

    """

    if len(newaxis_dims) == 0:
        return inp

    result_shape, reassoc_list = compute_result_shape_reassoc_list(inp.shape, newaxis_dims)

    return MemRefValue(
        memref.expand_shape(
            T.memref(*result_shape, inp.dtype),
            inp,
            reassoc_list,
            output_shape=[],
            static_output_shape=result_shape,
            loc=loc,
            ip=ip,
        )
    )


def _infer_memref_subview_result_type(source_memref_type, offsets, sizes, strides):
    source_strides, source_offset = source_memref_type.get_strides_and_offset()
    # "canonicalize" from tuple|list -> list
    offsets, sizes, strides, source_strides = map(list, (offsets, sizes, strides, source_strides))

    if any(not _is_static_int_like(i) for i in offsets + [source_offset]):
        target_offset = ShapedType.get_dynamic_size()
    else:
        target_offset = source_offset
        for offset, target_stride in zip(offsets, source_strides):
            target_offset += offset * target_stride

    target_strides = []
    for source_stride, static_stride in zip(source_strides, strides):
        target_strides.append(source_stride * static_stride)

    if all(isinstance(s, int) for s in sizes):
        # If default striding then no need to complicate things for downstream ops (e.g., expand_shape).
        default_strides = list(accumulate(sizes[1:][::-1], operator.mul))[::-1] + [1]
    else:
        default_strides = None
        sizes = [s if isinstance(s, int) else ShapedType.get_dynamic_size() for s in sizes]

    if target_strides == default_strides and target_offset == 0:
        layout = None
    else:
        layout = StridedLayoutAttr.get(target_offset, target_strides)

    return MemRefType.get(
        sizes,
        source_memref_type.element_type,
        layout,
        source_memref_type.memory_space,
    )


def subview(
    source: Value,
    offsets: MixedValues,
    sizes: MixedValues,
    strides: MixedValues,
    *,
    rank_reduce=False,
    result_type: Optional[MemRefType] = None,
    loc=None,
    ip=None,
):
    if offsets is None:
        offsets = []
    if sizes is None:
        sizes = []
    if strides is None:
        strides = []

    for s in [offsets, sizes, strides]:
        for idx, i in enumerate(s):
            if _is_constant_int_like(i):
                s[idx] = i.owner.opview.literal_value

    if result_type is None:
        # If any are arith.constant results then this will canonicalize to python int
        # (which can then be used to fully specify the subview).
        result_type = _infer_memref_subview_result_type(source.type, offsets, sizes, strides)

    offsets, _packed_offsets, static_offsets = _dispatch_mixed_values(offsets)
    sizes, _packed_sizes, static_sizes = _dispatch_mixed_values(sizes)
    strides, _packed_strides, static_strides = _dispatch_mixed_values(strides)

    if rank_reduce:
        result_shape = list(result_type.shape)
        layout_strides = None
        if result_type.layout:
            layout_strides = result_type.layout.strides
        for i, (s, ss) in reversed(list(enumerate(list(zip_longest(sizes, static_sizes))))):
            if (s is not None and _is_static_int_like(s) and s.literal_value == 1) or ss == 1:
                del result_shape[i]
                if layout_strides is not None:
                    del layout_strides[i]
        reduced_layout = None
        if layout_strides is not None:
            reduced_layout = StridedLayoutAttr.get(result_type.layout.offset, layout_strides)
        result_type = MemRefType.get(
            result_shape,
            result_type.element_type,
            reduced_layout,
            result_type.memory_space,
        )

    return _generated_subview(
        result_type,
        source,
        offsets,
        sizes,
        strides,
        static_offsets,
        static_sizes,
        static_strides,
        loc=loc,
        ip=ip,
    )


def _subview(
    mem: MemRefValue,
    idx,
    *,
    rank_reduce=False,
    loc=None,
    ip=None,
) -> MemRefValue:
    indexer = _indices_to_indexer(idx, mem.shape)
    out = mem

    if indexer.is_constant():
        offsets = indexer.static_offsets()
        sizes = indexer.static_sizes()
        strides = indexer.static_strides()
    else:
        # special tile case
        offsets = [None] * len(indexer.in_shape)
        sizes = [None] * len(indexer.in_shape)
        strides = [None] * len(indexer.in_shape)
        for i, ind in enumerate(indexer.indices):
            if isinstance(ind, slice):
                maybe_size = _maybe_compute_size(ind.start, ind.stop, ind.step)
                if maybe_size is None:
                    raise RuntimeError(f"failed to canonicalize start, stop, step: {ind=}")
                offsets[i] = ind.start
                sizes[i] = maybe_size
                strides[i] = (
                    ind.step.literal_value if isinstance(ind.step, ScalarValue) else ind.step
                )
            elif isinstance(ind, Value):
                offsets[i] = ind
                sizes[i] = 1
                strides[i] = 1
            else:
                raise RuntimeError(f"indexing of {mem=} not supported by {ind=}")
        assert all(map(lambda x: x is not None, offsets + sizes + strides)), (
            f"not each slice is statically known: {indexer.indices}"
        )

    out = subview(
        out,
        offsets=offsets,
        sizes=sizes,
        strides=strides,
        rank_reduce=rank_reduce,
        loc=loc,
        ip=ip,
    )

    # This adds newaxis/None dimensions.
    return expand_shape(out, indexer.newaxis_dims, loc=loc, ip=ip)


def _copy_to_subview(
    dest: MemRefValue,
    source: MemRefValue,
    idx,
    *,
    loc=None,
    ip=None,
):
    if isinstance(source, ScalarValue):
        source = expand_shape(source, (0,), loc=loc, ip=ip)

    dest_subview = _subview(dest, idx, loc=loc, ip=ip)
    assert dest_subview.shape == source.shape, (
        f"Expected matching shape for dest subview {dest_subview.shape} and source {source.shape=}"
    )

    return memref.copy(source, dest_subview, loc=loc, ip=ip)


alloca_scope = region_op(AllocaScopeOp)

_dim = dim


def dim(source, index, *, loc=None, ip=None):
    if isinstance(index, int):
        index = constant(index, index=True)
    return _dim(source=source, index=index, loc=loc, ip=ip)


def global_(
    initial_value=None,
    sym_name=None,
    type=None,
    sym_visibility="private",
    constant=None,
    alignment=None,
    loc=None,
    ip=None,
):
    if sym_name is None:
        previous_frame = inspect.currentframe().f_back
        sym_name = _get_sym_name(previous_frame, check_func_call="memref\\.global_|global_")
        assert sym_name is not None, "couldn't automatically find sym_name in previous frame"
    if initial_value is None:
        assert type is not None
    else:
        assert isinstance(initial_value, np.ndarray)
        if type is None:
            type = infer_mlir_type(initial_value, memref=True)
        initial_value = DenseElementsAttr.get(
            initial_value,
            type=type.element_type,
            context=None,
        )
        constant = True

    return memref.global_(
        sym_name,
        type,
        sym_visibility=sym_visibility,
        initial_value=initial_value,
        constant=constant,
        alignment=alignment,
        loc=loc,
        ip=ip,
    ).opview


def view(source, shape, dtype=None, shift=0, memory_space=None, loc=None, ip=None):
    if dtype is None:
        dtype = source.type.element_type
    byte_width_dtype = dtype.width // 8
    byte_shift = shift * byte_width_dtype
    if isinstance(byte_shift, int):
        byte_shift = constant(byte_shift, index=True)
    elif isinstance(byte_shift, Value):
        if not isinstance(byte_shift.type, IndexType):
            byte_shift = index_cast(byte_shift, to=IndexType.get(), loc=loc, ip=ip)
    else:
        raise TypeError(f"expected {byte_shift=} to be either int or Value")
    assert isinstance(byte_shift.type, IndexType), "expected index type for byte-shift"
    if memory_space is None and source:
        memory_space = source.type.memory_space

    dynamic_sizes = []
    memref_shape = []
    for s in shape:
        if isinstance(s, int):
            memref_shape.append(s)
        else:
            memref_shape.append(ShapedType.get_dynamic_size())
            if not isinstance(s.type, IndexType):
                s = index_cast(s, to=IndexType.get(), loc=loc, ip=ip)
            dynamic_sizes.append(s)

    return memref.view(
        T.memref(*memref_shape, element_type=dtype, memory_space=memory_space),
        source,
        byte_shift,
        dynamic_sizes,
        loc=loc,
        ip=ip,
    )


_get_global = get_global


def get_global(name_or_global, *, name=None, global_=None, result=None, loc=None, ip=None):
    if isinstance(name_or_global, GlobalOp):
        global_ = name_or_global
    elif isinstance(name_or_global, str):
        name = name_or_global
    elif name_or_global is not None:
        raise ValueError(f"only string or GlobalOp can be provided; got {name_or_global}")

    if global_ is None:
        assert name is not None, "name must be provided"

        if result is None:

            def callback(symbol_table_op, _uses_visible):
                nonlocal global_
                sym_table = SymbolTable(symbol_table_op)
                if name in sym_table:
                    global_ = sym_table[name]

            current_owner = InsertionPoint.current.block.owner
            while not isinstance(current_owner.opview, builtin.ModuleOp):
                current_owner = current_owner.parent
            SymbolTable.walk_symbol_tables(current_owner, True, callback)
            if global_ is None:
                raise RuntimeError(f"couldn't find symbol for {name}")

    if not isinstance(global_, GlobalOp):
        raise RuntimeError(f"expected memref.global, got {global_}")
    result = global_.type_.value
    name = global_.sym_name.value
    return GetGlobalOp(result=result, name=name, loc=loc, ip=ip).result


def reinterpret_cast(
    source: Value,
    offsets: MixedValues = None,
    sizes: MixedValues = None,
    strides: MixedValues = None,
    *,
    loc=None,
    ip=None,
) -> Value:
    if offsets is None:
        offsets = []
    if sizes is None:
        sizes = []

    offsets_, _packed_offsets, static_offsets = _dispatch_mixed_values(offsets)
    sizes_, _packed_sizes, static_sizes = _dispatch_mixed_values(sizes)
    strides_, _packed_strides, static_strides = _dispatch_mixed_values(strides)

    if offsets_ or sizes_ or strides_:
        raise NotImplementedError("only static offsets and sizes and strides supported")

    default_strides = None
    if not static_strides and all(_is_static_int_like(s) for s in static_sizes):
        default_strides = list(accumulate(list(static_sizes)[1:][::-1], operator.mul))[::-1] + [1]
        static_strides = default_strides

    target_offset = 0
    for offset, target_stride in zip(static_offsets, static_strides):
        target_offset += offset * target_stride

    if static_strides == default_strides and target_offset == 0:
        layout = None
    else:
        layout = StridedLayoutAttr.get(target_offset, static_strides)

    result = MemRefType.get(
        static_sizes, source.type.element_type, layout, source.type.memory_space
    )
    return ReinterpretCastOp(
        result=result,
        source=source,
        offsets=offsets_,
        sizes=sizes_,
        strides=strides_,
        static_offsets=static_offsets,
        static_sizes=static_sizes,
        static_strides=static_strides,
        loc=loc,
        ip=ip,
    ).result



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# AST transformation passes for numba_cuda_mlir
# These run before Numba's IR conversion
import ast
from typing import Callable

from numba_cuda_mlir.ast_transforms.common import get_function_ast, recompile_function
from numba_cuda_mlir.ast_transforms.comprehension import ComprehensionPass
from numba_cuda_mlir.ast_transforms.consteval import ConstevalError, ConstevalPass
from numba_cuda_mlir.ast_transforms.constant_if import ConstantIfPass
from numba_cuda_mlir.ast_transforms.pipeline import (
    ASTTransformPass,
    ASTTransformPipeline,
    TransformContext,
)

__all__ = [
    "apply_ast_transforms",
    "ConstevalError",
    "ASTTransformPass",
    "ASTTransformPipeline",
    "TransformContext",
]


class NoneStatementRemover(ast.NodeTransformer):
    """Remove bare None expression statements (e.g., from consteval(print(...)))."""

    def __init__(self):
        self.modified = False

    def visit_Expr(self, node):
        if isinstance(node.value, ast.Constant) and node.value.value is None:
            self.modified = True
            return None  # Remove the node
        return node


def remove_none_statements(tree: ast.AST) -> tuple[ast.AST, bool]:
    """Remove bare None statements from the AST."""
    remover = NoneStatementRemover()
    tree = remover.visit(tree)
    ast.fix_missing_locations(tree)
    return tree, remover.modified


class NoneStatementRemovalPass(ASTTransformPass):
    """Pipeline pass that removes bare None expression statements."""

    @property
    def name(self) -> str:
        return "NoneStatementRemoval"

    def transform(self, tree: ast.Module, context: TransformContext) -> tuple[ast.Module, bool]:
        return remove_none_statements(tree)


def create_default_pipeline() -> ASTTransformPipeline:
    """Create the default AST transformation pipeline."""
    pipeline = ASTTransformPipeline()
    pipeline.add_pass(ConstevalPass())
    pipeline.add_pass(ConstantIfPass())
    pipeline.add_pass(ComprehensionPass())
    pipeline.add_pass(NoneStatementRemovalPass())
    return pipeline


def apply_ast_transforms(
    func: Callable,
    targetoptions: dict = None,
    argtypes: tuple = None,
) -> tuple[Callable, str | None]:
    """Apply AST transformations to a function at compile time.

    Args:
        func: The function to transform
        targetoptions: Compilation options dict. Supports:
            - experimental_ast_transforms: Enable AST transforms (required)
            - dump_ast: Print AST before/after all transformations
            - dump_ast_after_all: Print AST after each transformation pass
        argtypes: Tuple of Numba types for the function arguments. When provided,
            parameter names in consteval expressions will resolve to their types.

    Returns tuple of (transformed_function, transformed_source).
    transformed_source is None if no transformations were applied.
    """
    targetoptions = targetoptions or {}

    # AST transforms are gated behind experimental flag
    if not targetoptions.get("experimental_ast_transforms", False):
        return func, None

    tree = get_function_ast(func)
    if tree is None:
        return func, None

    dump_ast = targetoptions.get("dump_ast", False)
    dump_ast_after_all = targetoptions.get("dump_ast_after_all", False)

    # Print original source if dump_ast is enabled (but not dump_ast_after_all,
    # which prints its own "before" output)
    if dump_ast and not dump_ast_after_all:
        print(f"=== AST for {func.__name__} (before transforms) ===")
        print(ast.unparse(tree))

    # Create context and pipeline
    context = TransformContext(
        func=func,
        targetoptions=targetoptions,
        argtypes=argtypes or (),
    )
    pipeline = create_default_pipeline()

    # Run the pipeline
    tree, modified = pipeline.run(tree, context, dump_after_all=dump_ast_after_all)

    transformed_source = None
    if modified:
        transformed_source = ast.unparse(tree)
        if dump_ast and not dump_ast_after_all:
            print(f"=== AST for {func.__name__} (after transforms) ===")
            if context.stored_values:
                for name, value in context.stored_values.items():
                    print(f"{name} = {value!r}")
            print(transformed_source)
        func = recompile_function(func, tree, context.stored_values)

    return func, transformed_source



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import sys
import math
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.typing.templates import (
    ConcreteTemplate,
    signature,
    Registry,
)

registry = Registry()
infer_global = registry.register_global


@infer_global(math.exp)
@infer_global(math.expm1)
@infer_global(math.fabs)
@infer_global(math.sqrt)
@infer_global(math.log)
@infer_global(math.log1p)
@infer_global(math.log10)
@infer_global(math.log2)
@infer_global(math.sin)
@infer_global(math.cos)
@infer_global(math.tan)
@infer_global(math.sinh)
@infer_global(math.cosh)
@infer_global(math.tanh)
@infer_global(math.asin)
@infer_global(math.acos)
@infer_global(math.atan)
@infer_global(math.asinh)
@infer_global(math.acosh)
@infer_global(math.atanh)
@infer_global(math.degrees)
@infer_global(math.radians)
@infer_global(math.erf)
@infer_global(math.erfc)
@infer_global(math.gamma)
@infer_global(math.lgamma)
class Math_unary(ConcreteTemplate):
    cases = [
        signature(types.float64, types.int64),
        signature(types.float64, types.uint64),
        signature(types.float32, types.float32),
        signature(types.float64, types.float64),
    ]


if sys.version_info >= (3, 11):
    Math_unary = infer_global(math.exp2)(Math_unary)


@infer_global(math.atan2)
class Math_atan2(ConcreteTemplate):
    cases = [
        signature(types.float64, types.int64, types.int64),
        signature(types.float64, types.uint64, types.uint64),
        signature(types.float32, types.float32, types.float32),
        signature(types.float64, types.float64, types.float64),
    ]


@infer_global(math.trunc)
class Math_converter(ConcreteTemplate):
    cases = [
        signature(types.intp, types.intp),
        signature(types.int64, types.int64),
        signature(types.uint64, types.uint64),
        signature(types.int64, types.float32),
        signature(types.int64, types.float64),
    ]


@infer_global(math.floor)
@infer_global(math.ceil)
class Math_floor_ceil(Math_converter):
    pass


@infer_global(math.copysign)
class Math_copysign(ConcreteTemplate):
    cases = [
        signature(types.float32, types.float32, types.float32),
        signature(types.float64, types.float64, types.float64),
    ]


@infer_global(math.hypot)
class Math_hypot(ConcreteTemplate):
    cases = [
        signature(types.float64, types.int64, types.int64),
        signature(types.float64, types.uint64, types.uint64),
        signature(types.float32, types.float32, types.float32),
        signature(types.float64, types.float64, types.float64),
    ]


@infer_global(math.nextafter)
class Math_nextafter(ConcreteTemplate):
    cases = [
        signature(types.float64, types.float64, types.float64),
        signature(types.float32, types.float32, types.float32),
    ]


@infer_global(math.isinf)
@infer_global(math.isnan)
class Math_predicate(ConcreteTemplate):
    cases = [
        signature(types.boolean, types.int64),
        signature(types.boolean, types.uint64),
        signature(types.boolean, types.float32),
        signature(types.boolean, types.float64),
    ]


@infer_global(math.isfinite)
class Math_isfinite(Math_predicate):
    pass


@infer_global(math.pow)
class Math_pow(ConcreteTemplate):
    cases = [
        signature(types.float64, types.float64, types.int64),
        signature(types.float64, types.float64, types.uint64),
        signature(types.float32, types.float32, types.float32),
        signature(types.float64, types.float64, types.float64),
    ]


@infer_global(math.gcd)
class Math_gcd(ConcreteTemplate):
    cases = [
        signature(types.int64, types.int64, types.int64),
        signature(types.int32, types.int32, types.int32),
        signature(types.int16, types.int16, types.int16),
        signature(types.int8, types.int8, types.int8),
        signature(types.uint64, types.uint64, types.uint64),
        signature(types.uint32, types.uint32, types.uint32),
        signature(types.uint16, types.uint16, types.uint16),
        signature(types.uint8, types.uint8, types.uint8),
    ]


@infer_global(math.frexp)
class Math_frexp(ConcreteTemplate):
    cases = [
        signature(types.Tuple((types.float64, types.intc)), types.float64),
        signature(types.Tuple((types.float32, types.intc)), types.float32),
    ]


@infer_global(math.ldexp)
class Math_ldexp(ConcreteTemplate):
    cases = [
        signature(types.float64, types.float64, types.intc),
        signature(types.float32, types.float32, types.intc),
    ]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from io import StringIO
import sys
from pathlib import Path
import black
from . import _get_all_libraries, _runtime
from textwrap import dedent


def stubgen(out: StringIO = sys.stdout) -> str:
    import numba_cuda_mlir

    unformatted = StringIO()

    numba_cuda_mlir_root = Path(numba_cuda_mlir.__path__[0]).parent
    runtime_path = Path(__file__).relative_to(numba_cuda_mlir_root)
    unformatted.write(
        dedent(
            f"""
    # Autogenerated by {runtime_path}
    # Please do not edit

    from numba_cuda_mlir.typing import none, ptr, int8, int16, int32, int64, bool, float32, float64, UniTuple, Tuple

    """
        )
    )
    for rtlib in _get_all_libraries():
        rtlib = getattr(_runtime, rtlib)
        for func in rtlib.functions.keys():
            unformatted.write(str(getattr(rtlib, func)) + "\n")
    unformatted.write("\n")
    out.write(black.format_str(unformatted.getvalue(), mode=black.Mode()))


if __name__ == "__main__":
    stubgen()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
This file provides internal compiler utilities that support certain special
operations with bytes and workarounds for limitations enforced in userland.
"""

from numba_cuda_mlir.numba_cuda.extending import intrinsic
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda import cgutils


# grab_byte / grab_uint64_t built llvmlite load codegen; filtered out on the MLIR
# path, so the codegen is a shared tombstone (typing retained).
def _dead_codegen(context, builder, signature, args):
    raise NotImplementedError("this byte-access codegen is not used on the MLIR path")


@intrinsic
def grab_byte(typingctx, data, offset):
    # returns a byte at a given offset in data
    sig = types.uint8(types.voidptr, types.intp)
    return sig, _dead_codegen


@intrinsic
def grab_uint64_t(typingctx, data, offset):
    # returns a uint64_t at a given offset in data
    sig = types.uint64(types.voidptr, types.intp)
    return sig, _dead_codegen


@intrinsic
def memcpy_region(typingctx, dst, dst_offset, src, src_offset, nbytes, align):
    """Copy nbytes from *(src + src_offset) to *(dst + dst_offset)"""

    def codegen(context, builder, signature, args):
        [
            dst_val,
            dst_offset_val,
            src_val,
            src_offset_val,
            nbytes_val,
            align_val,
        ] = args
        src_ptr = builder.gep(src_val, [src_offset_val])
        dst_ptr = builder.gep(dst_val, [dst_offset_val])
        cgutils.raw_memcpy(builder, dst_ptr, src_ptr, nbytes_val, align_val)
        return context.get_dummy_value()

    sig = types.void(
        types.voidptr,
        types.intp,
        types.voidptr,
        types.intp,
        types.intp,
        types.intp,
    )
    return sig, codegen



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Complex number intrinsics and cmath module support for numba-cuda-mlir.

This module provides:
1. Complex attribute operations (.real, .imag, .conjugate())
2. cmath module functions (phase, polar, rect, exp, log, sqrt, trig functions, etc.)
"""

import cmath
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir._mlir.dialects import (
    complex as complex_dialect,
    arith,
    math as math_dialect,
    scf,
    func,
)
from numba_cuda_mlir._mlir.extras import types as T
import numba_cuda_mlir._mlir.ir as ir
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
lower = registry.lower
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir.lowering_utilities import (
    convert,
    DeferredMethodCall,
    get_or_insert_function,
)
from numba_cuda_mlir.descriptor import MLIRTargetContext

# Constants used for overflow protection
FLT_MAX = 3.4028235e38
DBL_MAX = 1.7976931348623157e308
SQRT2 = 1.414213562373095048801688724209698079

# ============================================================================
# Built-in complex() constructor
# ============================================================================


@lower(complex, types.Number, types.Number)
@lower(types.NumberClass, types.Number, types.Number)
def complex_constructor_2args_cg(mlir_lower, target, args, kwargs):
    """Code generator for complex(real, imag) constructor."""
    assert not kwargs, "complex constructor does not accept keyword arguments"
    assert len(args) == 2, "complex constructor expects exactly 2 arguments"
    real = mlir_lower.load_var(args[0])
    imag = mlir_lower.load_var(args[1])

    # Create complex number from real and imag
    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)
    element_type = target_mlir_type.element_type
    # Convert to target element type
    real = convert(real, element_type)
    imag = convert(imag, element_type)
    result = complex_dialect.create_(target_mlir_type, real, imag)
    mlir_lower.store_var(target, result)


@lower(complex, types.Number)
def complex_constructor_1arg_cg(mlir_lower, target, args, kwargs):
    """Code generator for complex(real) constructor - imag defaults to 0."""
    assert not kwargs, "complex constructor does not accept keyword arguments"
    assert len(args) == 1, "complex constructor expects exactly 1 argument"
    real = mlir_lower.load_var(args[0])

    # Create complex number from real with imag=0
    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)
    element_type = target_mlir_type.element_type
    # Convert to target element type
    real = convert(real, element_type)
    zero = arith.constant(result=element_type, value=0.0)
    result = complex_dialect.create_(target_mlir_type, real, zero)
    mlir_lower.store_var(target, result)


@lower(complex)
def complex_constructor_0arg_cg(mlir_lower, target, args, kwargs):
    """Code generator for complex() constructor with no args - returns 0+0j."""
    assert not kwargs, "complex constructor does not accept keyword arguments"
    assert len(args) == 0, "complex constructor expects no arguments"

    # Create complex number 0+0j
    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)
    element_type = target_mlir_type.element_type
    zero = arith.constant(result=element_type, value=0.0)
    result = complex_dialect.create_(target_mlir_type, zero, zero)
    mlir_lower.store_var(target, result)


# ============================================================================
# Complex Attribute Operations
# ============================================================================


@registry.lower_getattr(types.Complex, "real")
def complex_real_getattr(context: MLIRTargetContext, builder, target, value):
    """Extract the real part of a complex number using complex.re operation."""
    trace("complex.real getattr")
    value_val = builder.load_var(value)
    target_type = builder.get_numba_type(target.name)
    target_mlir_type = builder.get_mlir_type(target_type)

    # Use MLIR's complex.re operation
    real_part = complex_dialect.re(value_val)
    real_part = convert(real_part, target_mlir_type)
    builder.store_var(target, real_part)


@registry.lower_getattr(types.Complex, "imag")
def complex_imag_getattr(context: MLIRTargetContext, builder, target, value):
    """Extract the imaginary part of a complex number using complex.im operation."""
    trace("complex.imag getattr")
    value_val = builder.load_var(value)
    target_type = builder.get_numba_type(target.name)
    target_mlir_type = builder.get_mlir_type(target_type)

    # Use MLIR's complex.im operation
    imag_part = complex_dialect.im(value_val)
    imag_part = convert(imag_part, target_mlir_type)
    builder.store_var(target, imag_part)


# For conjugate, we use DeferredMethodCall like Array methods
def complex_conjugate_cg(mlir_lower, target, args, kwargs):
    """Code generator for complex conjugate operation."""
    assert not kwargs, "conjugate does not accept any keyword arguments"
    assert len(args) == 1, "conjugate expects exactly 1 argument (self)"
    x = mlir_lower.load_var(args[0])
    result = complex_dialect.conj(x)
    mlir_lower.store_var(target, result)


@registry.lower_getattr(types.Complex, "conjugate")
def complex_conjugate_getattr(context: MLIRTargetContext, builder, target, value):
    """
    Handle x.conjugate attribute access.
    Returns a DeferredMethodCall that will execute the conjugate when called.
    """
    trace("complex.conjugate getattr")
    builder.store_var(target, DeferredMethodCall(value, complex_conjugate_cg))


# ============================================================================
# cmath Conversion Functions
# ============================================================================


@lower(cmath.phase, types.Complex)
def cmath_phase_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.phase."""
    assert not kwargs, "phase does not accept any keyword arguments"
    assert len(args) == 1, "phase expects exactly 1 argument"
    x = mlir_lower.load_var(args[0])
    # phase = atan2(imag, real)
    imag = complex_dialect.im(x)
    real = complex_dialect.re(x)
    result = math_dialect.atan2(imag, real)
    mlir_lower.store_var(target, result)


@lower(cmath.polar, types.Complex)
def cmath_polar_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.polar."""
    assert not kwargs, "polar does not accept any keyword arguments"
    assert len(args) == 1, "polar expects exactly 1 argument"
    x = mlir_lower.load_var(args[0])
    real = complex_dialect.re(x)
    imag = complex_dialect.im(x)
    # r = hypot(real, imag)
    r_sq = arith.addf(arith.mulf(real, real), arith.mulf(imag, imag))
    r = math_dialect.sqrt(r_sq)
    # phi = atan2(imag, real)
    phi = math_dialect.atan2(imag, real)
    mlir_lower.store_var(target, (r, phi))


@lower(cmath.rect, types.Number, types.Number)
def cmath_rect_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.rect with special case handling."""
    from numba_cuda_mlir.runtime import cmath as cmath_lib
    from numba_cuda_mlir.lowering_utilities import link

    assert not kwargs, "rect does not accept any keyword arguments"
    assert len(args) == 2, "rect expects exactly 2 arguments"
    r = mlir_lower.load_var(args[0])
    phi = mlir_lower.load_var(args[1])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)
    element_type = target_mlir_type.element_type
    type_suffix = "f32" if element_type == T.f32() else "f64"
    intrinsic_name = f"cmath_rect_type_{type_suffix}"

    # Link the cmath library
    if cmath_lib not in mlir_lower._seen_mlir_libraries:
        mlir_lower._seen_mlir_libraries.add(cmath_lib)
        link.link_inplace(mlir_lower.mlir_module, cmath_lib.source)

    # Ensure r and phi are the right type
    r = convert(r, element_type)
    phi = convert(phi, element_type)

    # Declare or lookup the function: (FTy, FTy) -> (FTy, FTy)
    fn_type = ir.FunctionType.get([element_type, element_type], [element_type, element_type])
    callee = get_or_insert_function(intrinsic_name, fn_type, mlir_lower.mlir_gpu_module)

    # Call the function
    call_results = func.call(
        result=[element_type, element_type],
        callee=callee.name.value,
        operands_=[r, phi],
    )

    # Create complex from results
    result = complex_dialect.create_(target_mlir_type, call_results[0], call_results[1])
    mlir_lower.store_var(target, result)


# ============================================================================
# cmath Classification Functions
# ============================================================================


@lower(cmath.isnan, types.Complex)
def cmath_isnan_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.isnan."""
    assert not kwargs, "isnan does not accept any keyword arguments"
    assert len(args) == 1, "isnan expects exactly 1 argument"
    x = mlir_lower.load_var(args[0])
    real = complex_dialect.re(x)
    imag = complex_dialect.im(x)
    real_isnan = math_dialect.isnan(real)
    imag_isnan = math_dialect.isnan(imag)
    result = arith.ori(real_isnan, imag_isnan)
    mlir_lower.store_var(target, result)


@lower(cmath.isinf, types.Complex)
def cmath_isinf_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.isinf."""
    assert not kwargs, "isinf does not accept any keyword arguments"
    assert len(args) == 1, "isinf expects exactly 1 argument"
    x = mlir_lower.load_var(args[0])
    real = complex_dialect.re(x)
    imag = complex_dialect.im(x)
    real_isinf = math_dialect.isinf(real)
    imag_isinf = math_dialect.isinf(imag)
    result = arith.ori(real_isinf, imag_isinf)
    mlir_lower.store_var(target, result)


@lower(cmath.isfinite, types.Complex)
def cmath_isfinite_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.isfinite."""
    assert not kwargs, "isfinite does not accept any keyword arguments"
    assert len(args) == 1, "isfinite expects exactly 1 argument"
    x = mlir_lower.load_var(args[0])
    real = complex_dialect.re(x)
    imag = complex_dialect.im(x)
    real_isfinite = math_dialect.isfinite(real)
    imag_isfinite = math_dialect.isfinite(imag)
    result = arith.andi(real_isfinite, imag_isfinite)
    mlir_lower.store_var(target, result)


# ============================================================================
# cmath Power and Logarithm Functions
# ============================================================================


def _call_cmath_intrinsic(mlir_lower, func_name, z, target_mlir_type):
    """Helper to call a cmath runtime intrinsic that takes (real, imag) and returns (real, imag)."""
    from numba_cuda_mlir.runtime import cmath as cmath_lib
    from numba_cuda_mlir.lowering_utilities import link

    element_type = target_mlir_type.element_type
    type_suffix = "f32" if element_type == T.f32() else "f64"
    intrinsic_name = f"{func_name}_type_{type_suffix}"

    # Link the cmath library
    if cmath_lib not in mlir_lower._seen_mlir_libraries:
        mlir_lower._seen_mlir_libraries.add(cmath_lib)
        link.link_inplace(mlir_lower.mlir_module, cmath_lib.source)

    x = convert(complex_dialect.re(z), element_type)
    y = convert(complex_dialect.im(z), element_type)

    # Lookup the function
    fn_type = ir.FunctionType.get([element_type, element_type], [element_type, element_type])
    callee = get_or_insert_function(intrinsic_name, fn_type, mlir_lower.mlir_gpu_module)

    # Call the function
    call_results = func.call(
        result=[element_type, element_type], callee=callee.name.value, operands_=[x, y]
    )

    # Create complex from results
    return complex_dialect.create_(target_mlir_type, call_results[0], call_results[1])


@lower(cmath.exp, types.Complex)
def cmath_exp_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.exp with proper special case handling."""
    assert not kwargs, "exp does not accept any keyword arguments"
    assert len(args) == 1, "exp expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_exp", z, target_mlir_type)
    mlir_lower.store_var(target, result)


@lower(cmath.log, types.Complex)
def cmath_log_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.log using log(z) = log|z| + i*arg(z)."""
    assert not kwargs, "log does not accept any keyword arguments"
    if len(args) == 1:
        z = mlir_lower.load_var(args[0])

        target_type = mlir_lower.get_numba_type(target.name)
        target_mlir_type = mlir_lower.get_mlir_type(target_type)

        x = complex_dialect.re(z)
        y = complex_dialect.im(z)

        # log|z| = log(hypot(x, y))
        x_sq = arith.mulf(x, x)
        y_sq = arith.mulf(y, y)
        hypot_xy = math_dialect.sqrt(arith.addf(x_sq, y_sq))
        log_abs = math_dialect.log(hypot_xy)

        # arg(z) = atan2(y, x)
        phase = math_dialect.atan2(y, x)

        result = complex_dialect.create_(target_mlir_type, log_abs, phase)
        mlir_lower.store_var(target, result)
    else:
        raise NotImplementedError("cmath.log with base not implemented")


@lower(cmath.log, types.Complex, types.Complex)
def cmath_log_base_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.log with base."""
    assert not kwargs, "log does not accept any keyword arguments"
    assert len(args) == 2, "log with base expects exactly 2 arguments"
    x = mlir_lower.load_var(args[0])
    base = mlir_lower.load_var(args[1])
    log_x = complex_dialect.log(x)
    log_base = complex_dialect.log(base)
    result = complex_dialect.div(log_x, log_base)
    mlir_lower.store_var(target, result)


@lower(cmath.log10, types.Complex)
def cmath_log10_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.log10 using log10(z) = log(z) / ln(10)."""
    assert not kwargs, "log10 does not accept any keyword arguments"
    assert len(args) == 1, "log10 expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)
    element_type = target_mlir_type.element_type

    x = convert(complex_dialect.re(z), element_type)
    y = convert(complex_dialect.im(z), element_type)

    # log10(z) = log(z) / ln(10), where log(z) = log|z| + i*arg(z)
    x_sq = arith.mulf(x, x)
    y_sq = arith.mulf(y, y)
    hypot_xy = math_dialect.sqrt(arith.addf(x_sq, y_sq))
    log_abs = math_dialect.log(hypot_xy)
    phase = math_dialect.atan2(y, x)

    ln10 = arith.constant(result=element_type, value=2.302585092994046)
    real = arith.divf(log_abs, ln10)
    imag = arith.divf(phase, ln10)

    result = complex_dialect.create_(target_mlir_type, real, imag)
    mlir_lower.store_var(target, result)


@lower(cmath.sqrt, types.Complex)
def cmath_sqrt_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.sqrt with proper special case handling."""
    assert not kwargs, "sqrt does not accept any keyword arguments"
    assert len(args) == 1, "sqrt expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_sqrt", z, target_mlir_type)
    mlir_lower.store_var(target, result)


# ============================================================================
# cmath Trigonometric Functions
# ============================================================================


@lower(cmath.cos, types.Complex)
def cmath_cos_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.cos with proper special case handling."""
    assert not kwargs, "cos does not accept any keyword arguments"
    assert len(args) == 1, "cos expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_cos", z, target_mlir_type)
    mlir_lower.store_var(target, result)


@lower(cmath.sin, types.Complex)
def cmath_sin_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.sin with proper special case handling."""
    assert not kwargs, "sin does not accept any keyword arguments"
    assert len(args) == 1, "sin expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_sin", z, target_mlir_type)
    mlir_lower.store_var(target, result)


@lower(cmath.tan, types.Complex)
def cmath_tan_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.tan."""
    assert not kwargs, "tan does not accept any keyword arguments"
    assert len(args) == 1, "tan expects exactly 1 argument"
    x = mlir_lower.load_var(args[0])
    result = complex_dialect.tan(x)
    mlir_lower.store_var(target, result)


@lower(cmath.acos, types.Complex)
def cmath_acos_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.acos using acos(z) = pi/2 - asin(z)."""
    assert not kwargs, "acos does not accept any keyword arguments"
    assert len(args) == 1, "acos expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)
    element_type = target_mlir_type.element_type

    asin_result = _call_cmath_intrinsic(mlir_lower, "cmath_asin", z, target_mlir_type)
    asin_real = complex_dialect.re(asin_result)
    asin_imag = complex_dialect.im(asin_result)

    pi_over_2 = arith.constant(result=element_type, value=1.5707963267948966)
    result_real = arith.subf(pi_over_2, asin_real)
    result_imag = arith.negf(asin_imag)
    result = complex_dialect.create_(target_mlir_type, result_real, result_imag)
    mlir_lower.store_var(target, result)


@lower(cmath.asin, types.Complex)
def cmath_asin_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.asin with proper special case handling."""
    assert not kwargs, "asin does not accept any keyword arguments"
    assert len(args) == 1, "asin expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_asin", z, target_mlir_type)
    mlir_lower.store_var(target, result)


@lower(cmath.atan, types.Complex)
def cmath_atan_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.atan with proper special case handling."""
    assert not kwargs, "atan does not accept any keyword arguments"
    assert len(args) == 1, "atan expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_atan", z, target_mlir_type)
    mlir_lower.store_var(target, result)


# ============================================================================
# cmath Hyperbolic Functions
# ============================================================================


@lower(cmath.cosh, types.Complex)
def cmath_cosh_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.cosh with proper special case handling."""
    assert not kwargs, "cosh does not accept any keyword arguments"
    assert len(args) == 1, "cosh expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_cosh", z, target_mlir_type)
    mlir_lower.store_var(target, result)


@lower(cmath.sinh, types.Complex)
def cmath_sinh_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.sinh with proper special case handling."""
    assert not kwargs, "sinh does not accept any keyword arguments"
    assert len(args) == 1, "sinh expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_sinh", z, target_mlir_type)
    mlir_lower.store_var(target, result)


@lower(cmath.tanh, types.Complex)
def cmath_tanh_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.tanh."""
    assert not kwargs, "tanh does not accept any keyword arguments"
    assert len(args) == 1, "tanh expects exactly 1 argument"
    x = mlir_lower.load_var(args[0])
    result = complex_dialect.tanh(x)
    mlir_lower.store_var(target, result)


@lower(cmath.acosh, types.Complex)
def cmath_acosh_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.acosh using acosh(z) = sign(z.imag)*i*acos(z).

    CPython: acosh(z) = i*acos(z) if z.imag >= 0, else -i*acos(z)
    Equivalently: acosh.real = |acos.imag|, acosh.imag = copysign(acos.real, z.imag)
    """
    assert not kwargs, "acosh does not accept any keyword arguments"
    assert len(args) == 1, "acosh expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)
    element_type = target_mlir_type.element_type

    # First compute asin(z), then acos = pi/2 - asin
    asin_result = _call_cmath_intrinsic(mlir_lower, "cmath_asin", z, target_mlir_type)
    asin_real = complex_dialect.re(asin_result)
    asin_imag = complex_dialect.im(asin_result)

    pi_over_2 = arith.constant(result=element_type, value=1.5707963267948966)
    acos_real = arith.subf(pi_over_2, asin_real)
    acos_imag = arith.negf(asin_imag)

    # acosh(z) = sign(z.imag) * i * acos(z)
    # i * (acos_real + i*acos_imag) = -acos_imag + i*acos_real
    # So: acosh.real = -acos_imag = |acos_imag| (should be non-negative)
    #     acosh.imag = copysign(acos_real, z.imag)
    z_imag = convert(complex_dialect.im(z), element_type)
    result_real = math_dialect.absf(acos_imag)
    result_imag = math_dialect.copysign(acos_real, z_imag)
    result = complex_dialect.create_(target_mlir_type, result_real, result_imag)
    mlir_lower.store_var(target, result)


@lower(cmath.asinh, types.Complex)
def cmath_asinh_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.asinh with proper special case handling."""
    assert not kwargs, "asinh does not accept any keyword arguments"
    assert len(args) == 1, "asinh expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)

    result = _call_cmath_intrinsic(mlir_lower, "cmath_asinh", z, target_mlir_type)
    mlir_lower.store_var(target, result)


@lower(cmath.atanh, types.Complex)
def cmath_atanh_cg(mlir_lower, target, args, kwargs):
    """Code generator for cmath.atanh using atanh(z) = 0.5 * log((1+z)/(1-z)).

    Special cases:
    - atanh(x ± inf*j) = copysign(0, x) ± pi/2*j for any x
    - For real z with |z| > 1: atanh(z) = atanh(1/z) + copysign(pi/2, imag(z))*j
    """
    assert not kwargs, "atanh does not accept any keyword arguments"
    assert len(args) == 1, "atanh expects exactly 1 argument"
    z = mlir_lower.load_var(args[0])

    target_type = mlir_lower.get_numba_type(target.name)
    target_mlir_type = mlir_lower.get_mlir_type(target_type)
    element_type = target_mlir_type.element_type

    one = arith.constant(result=element_type, value=1.0)
    half = arith.constant(result=element_type, value=0.5)
    zero = arith.constant(result=element_type, value=0.0)
    pi_over_2 = arith.constant(result=element_type, value=1.5707963267948966)
    one_complex = complex_dialect.create_(target_mlir_type, one, zero)
    half_complex = complex_dialect.create_(target_mlir_type, half, zero)

    x = complex_dialect.re(z)
    y = complex_dialect.im(z)

    # Special cases:
    # - imag is infinite: atanh(x ± inf*j) = copysign(0, x) ± pi/2*j
    # - real is infinite: atanh(±inf + y*j) = copysign(0, x) + copysign(pi/2, y)*j
    # - y is nan with x infinite: result = copysign(0, x) + nan*j
    y_is_inf = math_dialect.isinf(y)
    x_is_inf = math_dialect.isinf(x)
    y_is_nan = math_dialect.isnan(y)

    # Check if z is effectively real (y == 0) and |x| > 1 (branch cut case)
    y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, y, zero)
    abs_x = math_dialect.absf(x)
    abs_x_gt_1 = arith.cmpf(arith.CmpFPredicate.OGT, abs_x, one)
    is_branch_cut = arith.andi(y_is_zero, abs_x_gt_1)

    # 1 + z
    one_plus_z = complex_dialect.add(one_complex, z)

    # 1 - z
    one_minus_z = complex_dialect.sub(one_complex, z)

    # (1+z)/(1-z)
    ratio = complex_dialect.div(one_plus_z, one_minus_z)

    # log((1+z)/(1-z))
    log_term = complex_dialect.log(ratio)

    # 0.5 * log(...)
    formula_result = complex_dialect.mul(half_complex, log_term)

    formula_real = complex_dialect.re(formula_result)
    formula_imag = complex_dialect.im(formula_result)

    # For branch cut case: correct the imaginary part sign based on sign of y (even if ±0)
    # The imaginary part should be copysign(pi/2, y) when |x| > 1 and y == ±0
    branch_cut_imag = math_dialect.copysign(pi_over_2, y)
    corrected_formula_imag = arith.select(is_branch_cut, branch_cut_imag, formula_imag)

    # Special case results
    # For y_is_inf: atanh(x ± inf*j) = copysign(0, x) ± pi/2*j
    # For x_is_inf: atanh(±inf + y*j) = copysign(0, x) + copysign(pi/2, y)*j
    # For y_is_nan with x_is_inf: atanh(±inf + nan*j) = copysign(0, x) + nan*j
    signed_zero = math_dialect.copysign(zero, x)
    imag_sign = math_dialect.copysign(pi_over_2, y)
    nan_val = arith.constant(result=element_type, value=float("nan"))

    # When y is nan with infinite x, imag should be nan
    y_nan_with_inf_x = arith.andi(y_is_nan, x_is_inf)

    # Priority: y_is_inf > (y_nan_with_inf_x for imag) > x_is_inf > branch_cut > formula
    final_real = arith.select(
        y_is_inf, signed_zero, arith.select(x_is_inf, signed_zero, formula_real)
    )
    final_imag = arith.select(
        y_is_inf,
        imag_sign,
        arith.select(
            y_nan_with_inf_x,
            nan_val,
            arith.select(x_is_inf, imag_sign, corrected_formula_imag),
        ),
    )
    result = complex_dialect.create_(target_mlir_type, final_real, final_imag)

    mlir_lower.store_var(target, result)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir._mlir.dialects import func, gpu, builtin
from numba_cuda_mlir._mlir.passmanager import PassManager
import logging

from numba_cuda_mlir.lowering_utilities.type_conversions import to_numba_type
from numba_cuda_mlir.lowering_utilities import context


class DiscoverFunctionsPass:
    def __init__(self):
        self.__name__ = DiscoverFunctionsPass.__name__
        self.functions = dict()

    def recurse(self, op: ir.Operation, _pass: PassManager):
        for region in op.regions:
            for _block in region.blocks:
                for operation in region.blocks[0]:
                    self(operation, _pass)

    def __call__(self, op: ir.Operation, _pass: PassManager):
        logging.debug(f"DiscoverFunctionsPass: {op.name}")

        if getattr(op, "name", None) == "builtin.module":
            self.recurse(op, _pass)
            return

        match op:
            case func.FuncOp():
                self.functions[op.name.value] = to_numba_type(op.type)
            case gpu.GPUFuncOp():
                self.functions[op.name.value] = to_numba_type(op.function_type.value)
            case gpu.GPUModuleOp():
                self.recurse(op, _pass)


def discover_functions(module_str: str) -> dict[str, func.FuncOp]:
    """
    Discover all functions in the module.
    """
    with context.get_context():
        try:
            module = ir.Module.parse(module_str)
        except Exception as e:
            raise ValueError(f"Failed to parse MLIR module from string:\n{module_str}\n{e}")
        pm = PassManager()
        p = DiscoverFunctionsPass()
        pm.add(p)
        pm.run(module.operation)

    return p.functions



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# AST transformation pipeline infrastructure
import ast
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Callable


@dataclass
class TransformContext:
    """Shared context passed through the AST transform pipeline."""

    func: Callable
    targetoptions: dict = field(default_factory=dict)
    argtypes: tuple = ()
    stored_values: dict = field(default_factory=dict)


class ASTTransformPass(ABC):
    """Base class for AST transformation passes.

    Subclasses must implement:
    - name: Human-readable name of the pass
    - transform(tree, context): Apply the transformation
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Human-readable name of this transformation pass."""
        pass

    @abstractmethod
    def transform(self, tree: ast.Module, context: TransformContext) -> tuple[ast.Module, bool]:
        """Apply this transformation to the AST.

        Args:
            tree: The AST to transform
            context: Shared context with function info and state

        Returns:
            (transformed_tree, was_modified)
        """
        pass


class ASTTransformPipeline:
    """Pipeline that runs AST transformation passes in sequence."""

    def __init__(self, passes: list[ASTTransformPass] = None):
        self._passes = passes or []

    def add_pass(self, transform_pass: ASTTransformPass) -> "ASTTransformPipeline":
        """Add a pass to the pipeline."""
        self._passes.append(transform_pass)
        return self

    def run(
        self,
        tree: ast.Module,
        context: TransformContext,
        dump_after_all: bool = False,
    ) -> tuple[ast.Module, bool]:
        """Run all passes on the AST.

        Args:
            tree: The AST to transform
            context: Shared context with function info and state
            dump_after_all: If True, print source before and after each pass

        Returns:
            (transformed_tree, any_modified)
        """
        any_modified = False
        func_name = context.func.__name__

        if dump_after_all:
            print(f"# Unparsed source before transforms")
            print(ast.unparse(tree))

        for transform_pass in self._passes:
            tree, modified = transform_pass.transform(tree, context)
            any_modified |= modified

            if dump_after_all:
                print(f"# Unparsed source after {transform_pass.name}")
                print(ast.unparse(tree))

        return tree, any_modified



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
class UFuncRegistry:
    def __init__(self, name: str | None = None):
        self.registry = {}

    def __str__(self):
        return f"UFuncRegistry(registry={self.registry})"

    def __repr__(self):
        return str(self)

    def register(self, ufunc):
        def decorator(lowering):
            self.registry[ufunc] = lowering
            return lowering

        return decorator

    def get(self, ufunc):
        return self.registry.get(ufunc)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Typing support for struct and union field access
"""

from numba_cuda_mlir.numba_cuda.typing.templates import AttributeTemplate, Registry
from numba_cuda_mlir.numba_cuda.extending import typeof_impl
from numba_cuda_mlir.type_defs.aggregate_types import AggregateType, UnionType
from numba_cuda_mlir.cuda.experimental.struct import StructInstance, StructTypeWrapper
from numba_cuda_mlir.cuda.experimental.union import UnionInstance, UnionTypeWrapper
from numba_cuda_mlir.numba_cuda import types as numba_types

registry = Registry()


@registry.register_attr
class AggregateTypeAttributeTemplate(AttributeTemplate):
    """
    Resolve attribute access on struct types (AggregateType).

    This tells Numba's type inference system what type each field has,
    enabling getattr operations like `my_struct.field_name`.
    """

    key = AggregateType

    def generic_resolve(self, typ, attr):
        """
        Resolve the type of a field access on a struct.

        Args:
            typ: The AggregateType instance (the struct type)
            attr: The attribute name being accessed (field name)

        Returns:
            The type of the field, or None if the field doesn't exist
        """
        # Look up the field in the struct's field list
        # Skip padding fields (field_name is None)
        for field_name, field_type, *_ in typ.fields:
            if field_name is not None and field_name == attr:
                return field_type

        # Field not found - return None and Numba will raise an error
        return None


@registry.register_attr
class UnionTypeAttributeTemplate(AttributeTemplate):
    """
    Resolve attribute access on union types (UnionType).

    This tells Numba's type inference system what type each variant has,
    enabling getattr operations like `my_union.variant_name`.
    """

    key = UnionType

    def generic_resolve(self, typ, attr):
        """
        Resolve the type of a variant access on a union.

        Args:
            typ: The UnionType instance (the union type)
            attr: The attribute name being accessed (variant name)

        Returns:
            The type of the variant, or None if the variant doesn't exist
        """
        # Look up the variant in the union's variant list
        for variant_name, variant_type in typ.variants:
            if variant_name == attr:
                # Unwrap StructTypeWrapper to get the underlying AggregateType
                if isinstance(variant_type, StructTypeWrapper):
                    return variant_type._type
                return variant_type

        # Variant not found - return None and Numba will raise an error
        return None


@typeof_impl.register(StructInstance)
def typeof_struct_instance(val, c):
    """
    Type inference for StructInstance objects.

    This allows passing struct instances created on the host to JIT-compiled kernels.
    """
    return val._struct_type


@typeof_impl.register(StructTypeWrapper)
def typeof_struct_type_wrapper(val, c):
    """
    Type inference for StructTypeWrapper.

    When used as a value (e.g., called in device code), we need to return
    a Function type with the registered template so the typing system can
    resolve calls to it.
    """

    # If used in device code (has a constructor template), return Function type
    if val._constructor_template is not None:
        return numba_types.Function(template=val._constructor_template)

    # Otherwise (used as type annotation), return the underlying type
    return val._type


@typeof_impl.register(UnionInstance)
def typeof_union_instance(val, c):
    """
    Type inference for UnionInstance objects.

    This allows passing union instances created on the host to JIT-compiled kernels.
    """
    return val._union_type


@typeof_impl.register(UnionTypeWrapper)
def typeof_union_type_wrapper(val, c):
    """
    Type inference for UnionTypeWrapper.

    When used as a value (e.g., called in device code), we need to return
    a Function type with the registered template so the typing system can
    resolve calls to it.
    """
    # If used in device code (has a constructor template), return Function type
    if val._constructor_template is not None:
        return numba_types.Function(template=val._constructor_template)

    # Otherwise (used as type annotation), return the underlying type
    return val._type



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import operator

from numba_cuda_mlir._mlir.dialects import arith, math as mlir_math
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.np import npdatetime_helpers

from numba_cuda_mlir.lowering_utilities import convert
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
lower = registry.lower

# datetime64 and timedelta64 are i64 under the hood, so all ops are integer ops.
# When operand units differ we scale to the result unit before operating.


def _scale_to_unit(val, src_unit, dest_unit):
    """Multiply an i64 value by the conversion factor from src_unit to dest_unit."""
    if src_unit == dest_unit or src_unit == "" or dest_unit == "":
        return val
    factor = npdatetime_helpers.get_timedelta_conversion_factor(src_unit, dest_unit)
    if factor is None or factor == 1:
        return val
    return arith.muli(val, arith.constant(T.i64(), factor))


def _load_scaled(builder, var, dest_unit):
    """Load a datetime/timedelta variable and scale it to dest_unit."""
    val = builder.load_var(var)
    src_unit = builder.get_numba_type(var).unit
    return _scale_to_unit(val, src_unit, dest_unit)


@lower(operator.sub, types.NPDatetime, types.NPDatetime)
@lower(operator.sub, types.NPDatetime, types.NPTimedelta)
@lower(operator.sub, types.NPTimedelta, types.NPTimedelta)
def datetime_sub(builder, target, args, kwargs):
    dest_unit = builder.get_numba_type(target).unit
    lhs = _load_scaled(builder, args[0], dest_unit)
    rhs = _load_scaled(builder, args[1], dest_unit)
    builder.store_var(target, arith.subi(lhs, rhs))


@lower(operator.add, types.NPDatetime, types.NPTimedelta)
@lower(operator.add, types.NPTimedelta, types.NPDatetime)
@lower(operator.add, types.NPTimedelta, types.NPTimedelta)
def datetime_add(builder, target, args, kwargs):
    dest_unit = builder.get_numba_type(target).unit
    lhs = _load_scaled(builder, args[0], dest_unit)
    rhs = _load_scaled(builder, args[1], dest_unit)
    builder.store_var(target, arith.addi(lhs, rhs))


def _datetime_cmp(predicate):
    def impl(builder, target, args, kwargs):
        lty = builder.get_numba_type(args[0])
        rty = builder.get_numba_type(args[1])
        common = npdatetime_helpers.get_best_unit(lty.unit, rty.unit)
        lhs = _load_scaled(builder, args[0], common)
        rhs = _load_scaled(builder, args[1], common)
        builder.store_var(target, arith.cmpi(predicate, lhs, rhs))

    return impl


@lower(operator.eq, types.NPDatetime, types.NPDatetime)
@lower(operator.eq, types.NPTimedelta, types.NPTimedelta)
def datetime_eq(builder, target, args, kwargs):
    return _datetime_cmp(arith.CmpIPredicate.eq)(builder, target, args, kwargs)


@lower(operator.ne, types.NPDatetime, types.NPDatetime)
@lower(operator.ne, types.NPTimedelta, types.NPTimedelta)
def datetime_ne(builder, target, args, kwargs):
    return _datetime_cmp(arith.CmpIPredicate.ne)(builder, target, args, kwargs)


@lower(operator.lt, types.NPDatetime, types.NPDatetime)
@lower(operator.lt, types.NPTimedelta, types.NPTimedelta)
def datetime_lt(builder, target, args, kwargs):
    return _datetime_cmp(arith.CmpIPredicate.slt)(builder, target, args, kwargs)


@lower(operator.le, types.NPDatetime, types.NPDatetime)
@lower(operator.le, types.NPTimedelta, types.NPTimedelta)
def datetime_le(builder, target, args, kwargs):
    return _datetime_cmp(arith.CmpIPredicate.sle)(builder, target, args, kwargs)


@lower(operator.gt, types.NPDatetime, types.NPDatetime)
@lower(operator.gt, types.NPTimedelta, types.NPTimedelta)
def datetime_gt(builder, target, args, kwargs):
    return _datetime_cmp(arith.CmpIPredicate.sgt)(builder, target, args, kwargs)


@lower(operator.ge, types.NPDatetime, types.NPDatetime)
@lower(operator.ge, types.NPTimedelta, types.NPTimedelta)
def datetime_ge(builder, target, args, kwargs):
    return _datetime_cmp(arith.CmpIPredicate.sge)(builder, target, args, kwargs)


# --- timedelta * {int, float} and {int, float} * timedelta ---


def _td_mul(builder, target, td_var, other_var):
    td = builder.load_var(td_var)
    other = builder.load_var(other_var)
    other_ty = builder.get_numba_type(other_var)
    if isinstance(other_ty, types.Float):
        prod = arith.mulf(convert(td, T.f64()), other)
        builder.store_var(target, convert(prod, T.i64()))
    else:
        builder.store_var(target, arith.muli(td, convert(other, T.i64())))


@lower(operator.mul, types.NPTimedelta, types.Integer)
@lower(operator.mul, types.NPTimedelta, types.Float)
def timedelta_mul(builder, target, args, kwargs):
    _td_mul(builder, target, args[0], args[1])


@lower(operator.mul, types.Integer, types.NPTimedelta)
@lower(operator.mul, types.Float, types.NPTimedelta)
def timedelta_rmul(builder, target, args, kwargs):
    _td_mul(builder, target, args[1], args[0])


# --- timedelta / {int, float} and timedelta // {int, float} ---


def _td_div_scalar(builder, target, td_var, other_var):
    td = builder.load_var(td_var)
    other = builder.load_var(other_var)
    other_ty = builder.get_numba_type(other_var)
    if isinstance(other_ty, types.Float):
        quot = arith.divf(convert(td, T.f64()), other)
        builder.store_var(target, convert(quot, T.i64()))
    else:
        builder.store_var(target, arith.divsi(td, convert(other, T.i64())))


@lower(operator.truediv, types.NPTimedelta, types.Integer)
@lower(operator.truediv, types.NPTimedelta, types.Float)
@lower(operator.floordiv, types.NPTimedelta, types.Integer)
@lower(operator.floordiv, types.NPTimedelta, types.Float)
def timedelta_div_scalar(builder, target, args, kwargs):
    _td_div_scalar(builder, target, args[0], args[1])


# --- timedelta / timedelta -> float64 ---


@lower(operator.truediv, types.NPTimedelta, types.NPTimedelta)
def timedelta_div_timedelta(builder, target, args, kwargs):
    lty = builder.get_numba_type(args[0])
    rty = builder.get_numba_type(args[1])
    common = npdatetime_helpers.get_best_unit(lty.unit, rty.unit)
    lhs = convert(_load_scaled(builder, args[0], common), T.f64())
    rhs = convert(_load_scaled(builder, args[1], common), T.f64())
    builder.store_var(target, arith.divf(lhs, rhs))


@lower(operator.floordiv, types.NPTimedelta, types.NPTimedelta)
def timedelta_floordiv_timedelta(builder, target, args, kwargs):
    lty = builder.get_numba_type(args[0])
    rty = builder.get_numba_type(args[1])
    common = npdatetime_helpers.get_best_unit(lty.unit, rty.unit)
    lhs = convert(_load_scaled(builder, args[0], common), T.f64())
    rhs = convert(_load_scaled(builder, args[1], common), T.f64())
    quot = arith.divf(lhs, rhs)
    builder.store_var(target, convert(arith.floorf(quot), T.i64()))


# --- unary: -timedelta, +timedelta, abs(timedelta) ---


@lower(operator.neg, types.NPTimedelta)
def timedelta_neg(builder, target, args, kwargs):
    val = builder.load_var(args[0])
    zero = arith.constant(T.i64(), 0)
    builder.store_var(target, arith.subi(zero, val))


@lower(operator.pos, types.NPTimedelta)
def timedelta_pos(builder, target, args, kwargs):
    builder.store_var(target, builder.load_var(args[0]))


@lower(abs, types.NPTimedelta)
def timedelta_abs(builder, target, args, kwargs):
    builder.store_var(target, mlir_math.absi(builder.load_var(args[0])))




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "index"

@_ods_cext.register_operation(_Dialect)
class AddOp(_ods_ir.OpView):
  r"""
  The `index.add` operation takes two index values and computes their sum.
  
  Example:
  
  ```mlir
  // c = a + b
  %c = index.add %a, %b
  ```
  """

  OPERATION_NAME = "index.add"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(AddOp)
class AddOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.add"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def add(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return AddOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class AndOp(_ods_ir.OpView):
  r"""
  The `index.and` operation takes two index values and computes their bitwise
  and.
  
  Example:
  
  ```mlir
  // c = a & b
  %c = index.and %a, %b
  ```
  """

  OPERATION_NAME = "index.and"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(AndOp)
class AndOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.and"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def and_(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return AndOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class BoolConstantOp(_ods_ir.OpView):
  r"""
  The `index.bool.constant` operation produces an bool-typed SSA value equal
  to either `true` or `false`.
  
  This operation is used to materialize bool constants that arise when folding
  `index.cmp`.
  
  Example:
  
  ```mlir
  %0 = index.bool.constant true
  ```
  """

  OPERATION_NAME = "index.bool.constant"

  _ODS_REGIONS = (0, True)

  def __init__(self, value: _Union[bool, _ods_ir.BoolAttr], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["value"] = (value if (
    isinstance(value, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('BoolAttr')) else
      _ods_ir.AttrBuilder.get('BoolAttr')(value, context=_ods_context))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def value(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["value"]

  @value.setter
  def value(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["value"] = value

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IntegerType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(BoolConstantOp)
class BoolConstantOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.bool.constant"

  @builtins.property
  def value(self) -> _ods_ir.BoolAttr:
    return self.attributes["value"]

def bool_constant(value: _Union[bool, _ods_ir.BoolAttr], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IntegerType]:
  return BoolConstantOp(value=value, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CastSOp(_ods_ir.OpView):
  r"""
  The `index.casts` operation enables conversions between values of index type
  and concrete fixed-width integer types. If casting to a wider integer, the
  value is sign-extended. If casting to a narrower integer, the value is
  truncated.
  
  Example:
  
  ```mlir
  // Cast to i32
  %0 = index.casts %a : index to i32
  
  // Cast from i64
  %1 = index.casts %b : i64 to index
  ```
  """

  OPERATION_NAME = "index.casts"

  _ODS_REGIONS = (0, True)

  def __init__(self, output: _ods_ir.Type, input: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(input)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(output)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def input(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CastSOp)
class CastSOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.casts"

  @builtins.property
  def input(self) -> _ods_ir.Value:
    return self.operands[0]

def casts(output: _ods_ir.Type, input: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CastSOp(output=output, input=input, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CastUOp(_ods_ir.OpView):
  r"""
  The `index.castu` operation enables conversions between values of index type
  and concrete fixed-width integer types. If casting to a wider integer, the
  value is zero-extended. If casting to a narrower integer, the value is
  truncated.
  
  Example:
  
  ```mlir
  // Cast to i32
  %0 = index.castu %a : index to i32
  
  // Cast from i64
  %1 = index.castu %b : i64 to index
  ```
  """

  OPERATION_NAME = "index.castu"

  _ODS_REGIONS = (0, True)

  def __init__(self, output: _ods_ir.Type, input: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(input)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(output)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def input(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def output(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CastUOp)
class CastUOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.castu"

  @builtins.property
  def input(self) -> _ods_ir.Value:
    return self.operands[0]

def castu(output: _ods_ir.Type, input: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CastUOp(output=output, input=input, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CeilDivSOp(_ods_ir.OpView):
  r"""
  The `index.ceildivs` operation takes two index values and computes their
  signed quotient. Treats the leading bit as the sign and rounds towards
  positive infinity, i.e. `7 / -2 = -3`.
  
  Note: division by zero and signed division overflow are undefined behaviour.
  
  Example:
  
  ```mlir
  // c = ceil(a / b)
  %c = index.ceildivs %a, %b
  ```
  """

  OPERATION_NAME = "index.ceildivs"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CeilDivSOp)
class CeilDivSOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.ceildivs"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def ceildivs(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return CeilDivSOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CeilDivUOp(_ods_ir.OpView):
  r"""
  The `index.ceildivu` operation takes two index values and computes their
  unsigned quotient. Treats the leading bit as the most significant and rounds
  towards positive infinity, i.e. `6 / -2 = 1`.
  
  Note: division by zero is undefined behaviour.
  
  Example:
  
  ```mlir
  // c = ceil(a / b)
  %c = index.ceildivu %a, %b
  ```
  """

  OPERATION_NAME = "index.ceildivu"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CeilDivUOp)
class CeilDivUOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.ceildivu"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def ceildivu(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return CeilDivUOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CmpOp(_ods_ir.OpView):
  r"""
  The `index.cmp` operation takes two index values and compares them according
  to the comparison predicate and returns an `i1`. The following comparisons
  are supported:
  
  -   `eq`:  equal
  -   `ne`:  not equal
  -   `slt`: signed less than
  -   `sle`: signed less than or equal
  -   `sgt`: signed greater than
  -   `sge`: signed greater than or equal
  -   `ult`: unsigned less than
  -   `ule`: unsigned less than or equal
  -   `ugt`: unsigned greater than
  -   `uge`: unsigned greater than or equal
  
  The result is `1` if the comparison is true and `0` otherwise.
  
  Example:
  
  ```mlir
  // Signed less than comparison.
  %0 = index.cmp slt(%a, %b)
  
  // Unsigned greater than or equal comparison.
  %1 = index.cmp uge(%a, %b)
  
  // Not equal comparison.
  %2 = index.cmp ne(%a, %b)
  ```
  """

  OPERATION_NAME = "index.cmp"

  _ODS_REGIONS = (0, True)

  def __init__(self, pred: _Union[_Any, _ods_ir.Attribute], lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["pred"] = (pred if (
    isinstance(pred, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('index.IndexCmpPredicateAttr')) else
      _ods_ir.AttrBuilder.get('index.IndexCmpPredicateAttr')(pred, context=_ods_context))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def pred(self) -> _ods_ir.Attribute:
    return self.operation.attributes["pred"]

  @pred.setter
  def pred(self, value: _ods_ir.Attribute):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["pred"] = value

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IntegerType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CmpOp)
class CmpOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.cmp"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

  @builtins.property
  def pred(self) -> _ods_ir.Attribute:
    return self.attributes["pred"]

def cmp(pred: _Union[_Any, _ods_ir.Attribute], lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IntegerType]:
  return CmpOp(pred=pred, lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class ConstantOp(_ods_ir.OpView):
  r"""
  The `index.constant` operation produces an index-typed SSA value equal to
  some index-typed integer constant.
  
  Example:
  
  ```mlir
  %0 = index.constant 42
  ```
  """

  OPERATION_NAME = "index.constant"

  _ODS_REGIONS = (0, True)

  def __init__(self, value: _Union[int, _ods_ir.IntegerAttr], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["value"] = (value if (
    isinstance(value, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('IndexAttr')) else
      _ods_ir.AttrBuilder.get('IndexAttr')(value, context=_ods_context))
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def value(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["value"]

  @value.setter
  def value(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["value"] = value

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(ConstantOp)
class ConstantOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.constant"

  @builtins.property
  def value(self) -> _ods_ir.IntegerAttr:
    return self.attributes["value"]

def constant(value: _Union[int, _ods_ir.IntegerAttr], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return ConstantOp(value=value, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class DivSOp(_ods_ir.OpView):
  r"""
  The `index.divs` operation takes two index values and computes their signed
  quotient. Treats the leading bit as the sign and rounds towards zero, i.e.
  `6 / -2 = -3`.
  
  Note: division by zero and signed division overflow are undefined behaviour.
  
  Example:
  
  ```mlir
  // c = a / b
  %c = index.divs %a, %b
  ```
  """

  OPERATION_NAME = "index.divs"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(DivSOp)
class DivSOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.divs"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def divs(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return DivSOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class DivUOp(_ods_ir.OpView):
  r"""
  The `index.divu` operation takes two index values and computes their
  unsigned quotient. Treats the leading bit as the most significant and rounds
  towards zero, i.e. `6 / -2 = 0`.
  
  Note: division by zero is undefined behaviour.
  
  Example:
  
  ```mlir
  // c = a / b
  %c = index.divu %a, %b
  ```
  """

  OPERATION_NAME = "index.divu"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(DivUOp)
class DivUOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.divu"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def divu(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return DivUOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class FloorDivSOp(_ods_ir.OpView):
  r"""
  The `index.floordivs` operation takes two index values and computes their
  signed quotient. Treats the leading bit as the sign and rounds towards
  negative infinity, i.e. `5 / -2 = -3`.
  
  Note: division by zero and signed division overflow are undefined behaviour.
  
  Example:
  
  ```mlir
  // c = floor(a / b)
  %c = index.floordivs %a, %b
  ```
  """

  OPERATION_NAME = "index.floordivs"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(FloorDivSOp)
class FloorDivSOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.floordivs"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def floordivs(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return FloorDivSOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MaxSOp(_ods_ir.OpView):
  r"""
  The `index.maxs` operation takes two index values and computes their signed
  maximum value. Treats the leading bit as the sign, i.e. `max(-2, 6) = 6`.
  
  Example:
  
  ```mlir
  // c = max(a, b)
  %c = index.maxs %a, %b
  ```
  """

  OPERATION_NAME = "index.maxs"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MaxSOp)
class MaxSOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.maxs"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def maxs(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return MaxSOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MaxUOp(_ods_ir.OpView):
  r"""
  The `index.maxu` operation takes two index values and computes their
  unsigned maximum value. Treats the leading bit as the most significant, i.e.
  `max(15, 6) = 15` or `max(-2, 6) = -2`.
  
  Example:
  
  ```mlir
  // c = max(a, b)
  %c = index.maxu %a, %b
  ```
  """

  OPERATION_NAME = "index.maxu"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MaxUOp)
class MaxUOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.maxu"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def maxu(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return MaxUOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MinSOp(_ods_ir.OpView):
  r"""
  The `index.mins` operation takes two index values and computes their signed
  minimum value. Treats the leading bit as the sign, i.e. `min(-2, 6) = -2`.
  
  Example:
  
  ```mlir
  // c = min(a, b)
  %c = index.mins %a, %b
  ```
  """

  OPERATION_NAME = "index.mins"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MinSOp)
class MinSOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.mins"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def mins(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return MinSOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MinUOp(_ods_ir.OpView):
  r"""
  The `index.minu` operation takes two index values and computes their
  unsigned minimum value. Treats the leading bit as the most significant, i.e.
  `min(15, 6) = 6` or `min(-2, 6) = 6`.
  
  Example:
  
  ```mlir
  // c = min(a, b)
  %c = index.minu %a, %b
  ```
  """

  OPERATION_NAME = "index.minu"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MinUOp)
class MinUOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.minu"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def minu(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return MinUOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MulOp(_ods_ir.OpView):
  r"""
  The `index.mul` operation takes two index values and computes their product.
  
  Example:
  
  ```mlir
  // c = a * b
  %c = index.mul %a, %b
  ```
  """

  OPERATION_NAME = "index.mul"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(MulOp)
class MulOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.mul"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def mul(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return MulOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class OrOp(_ods_ir.OpView):
  r"""
  The `index.or` operation takes two index values and computes their bitwise
  or.
  
  Example:
  
  ```mlir
  // c = a | b
  %c = index.or %a, %b
  ```
  """

  OPERATION_NAME = "index.or"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(OrOp)
class OrOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.or"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def or_(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return OrOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RemSOp(_ods_ir.OpView):
  r"""
  The `index.rems` operation takes two index values and computes their signed
  remainder. Treats the leading bit as the sign, i.e. `6 % -2 = 0`.
  
  Example:
  
  ```mlir
  // c = a % b
  %c = index.rems %a, %b
  ```
  """

  OPERATION_NAME = "index.rems"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RemSOp)
class RemSOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.rems"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def rems(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return RemSOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RemUOp(_ods_ir.OpView):
  r"""
  The `index.remu` operation takes two index values and computes their
  unsigned remainder. Treats the leading bit as the most significant, i.e.
  `6 % -2 = 6`.
  
  Example:
  
  ```mlir
  // c = a % b
  %c = index.remu %a, %b
  ```
  """

  OPERATION_NAME = "index.remu"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RemUOp)
class RemUOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.remu"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def remu(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return RemUOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class ShlOp(_ods_ir.OpView):
  r"""
  The `index.shl` operation shifts an index value to the left by a variable
  amount. The low order bits are filled with zeroes. The RHS operand is always
  treated as unsigned. If the RHS operand is equal to or greater than the
  index bitwidth, the result is a poison value.
  
  Example:
  
  ```mlir
  // c = a << b
  %c = index.shl %a, %b
  ```
  """

  OPERATION_NAME = "index.shl"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(ShlOp)
class ShlOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.shl"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def shl(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return ShlOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class ShrSOp(_ods_ir.OpView):
  r"""
  The `index.shrs` operation shifts an index value to the right by a variable
  amount. The LHS operand is treated as signed. The high order bits are filled
  with copies of the most significant bit. If the RHS operand is equal to or
  greater than the index bitwidth, the result is a poison value.
  
  Example:
  
  ```mlir
  // c = a >> b
  %c = index.shrs %a, %b
  ```
  """

  OPERATION_NAME = "index.shrs"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(ShrSOp)
class ShrSOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.shrs"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def shrs(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return ShrSOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class ShrUOp(_ods_ir.OpView):
  r"""
  The `index.shru` operation shifts an index value to the right by a variable
  amount. The LHS operand is treated as unsigned. The high order bits are
  filled with zeroes. If the RHS operand is equal to or greater than the index
  bitwidth, the result is a poison value.
  
  Example:
  
  ```mlir
  // c = a >> b
  %c = index.shru %a, %b
  ```
  """

  OPERATION_NAME = "index.shru"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(ShrUOp)
class ShrUOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.shru"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def shru(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return ShrUOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class SizeOfOp(_ods_ir.OpView):
  r"""
  The `index.sizeof` operation produces an index-typed SSA value equal to the
  size in bits of the `index` type. For example, on 32-bit systems, the result
  is `32 : index`, and on 64-bit systems, the result is `64 : index`.
  
  Example:
  
  ```mlir
  %0 = index.sizeof
  ```
  """

  OPERATION_NAME = "index.sizeof"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(SizeOfOp)
class SizeOfOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.sizeof"

def sizeof(*, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return SizeOfOp(results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class SubOp(_ods_ir.OpView):
  r"""
  The `index.sub` operation takes two index values and computes the difference
  of the first from the second operand.
  
  Example:
  
  ```mlir
  // c = a - b
  %c = index.sub %a, %b
  ```
  """

  OPERATION_NAME = "index.sub"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(SubOp)
class SubOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.sub"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def sub(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return SubOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class XOrOp(_ods_ir.OpView):
  r"""
  The `index.xor` operation takes two index values and computes their bitwise
  xor.
  
  Example:
  
  ```mlir
  // c = a ^ b
  %c = index.xor %a, %b
  ```
  """

  OPERATION_NAME = "index.xor"

  _ODS_REGIONS = (0, True)

  def __init__(self, lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(lhs)
    operands.append(rhs)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[1]

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(XOrOp)
class XOrOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "index.xor"

  @builtins.property
  def lhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

  @builtins.property
  def rhs(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[1]

def xor(lhs: _ods_ir.Value[_ods_ir.IndexType], rhs: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return XOrOp(lhs=lhs, rhs=rhs, results=results, loc=loc, ip=ip).result
